/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.lsf;

import org.apache.log4j.Logger;

import org.glite.apel.core.Log;
import org.glite.apel.core.Time;
import org.glite.apel.core.TimeInterval;
import org.glite.apel.core.log.type.EventRecord;
import org.glite.apel.core.log.type.EventRecordIterator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;

import java.text.SimpleDateFormat;

import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

/**
 * Parses LSF type log files, converting valid entries into EventRecords.
 */
class EventLogParser extends File {

    /**
     * Using log4j logging tool.
     */
    protected static final Logger LOG = Logger.getLogger(EventLogParser.class.getName());
    /**
     * The format used to encode the current date.
     */
    private static final String EVENT_DATE_FORMAT = "yyyyMMdd";
    /**
     * UTC constant used when converting the current date.
     */
    private static final String UTC = "UTC";
    /**
     * Job exit status constant.
     * 32 = EXIT (64=DONE)
     */
    private static final String JOB_EXIT_STATUS = "32";
    /**
     * Buffer which reads-in data from the log file.
     */
    protected BufferedReader m_reader;
    /**
     * Holds all event records read-in from the event log file.
     */
    protected Vector m_eventRecords;
    /**
     * The timezone to be used when interpreting time stamps
     * parsed from the log files.
     */
    private String m_timezone;
    /**
     * Pattern used to obtain the version of LSF
     */
    private Pattern m_delimiter_init;
    /**
     * Pattern used to chop up a line of data that matches an
     * EventRecord into a convenient list.
     */
    private Pattern m_delimiter;
    /**
     * The value for submitHost taken from the configuration file
     */
    private String m_submitHost;
    /**
     * Savannah bug #85096: Whether to scale CPU and wallclock time 
     * according to the value in the log file.  (If not, a scaling factor
     * must be set in the site BDII / GIIS which scales during the 
     * 'join' process of the publisher.)
     */
    private boolean m_useScalingFactor;

    /**
     * The file object representing this log file
     */
    private File m_file;
    /**
     * The values that will be used to create an EventRecords record
     * taken from the log file
     */
    private String eventID = "";
    private String jobName = "";
    private String LocalUserID = "";
    private String LocalUserGroup = "";
    private String wallDuration = "";
    private String cpuDuration = "";
    private String startTime = "";
    private String stopTime = "";
    private String hostFactor = "";
    private String memoryReal = "";
    private String memoryVirtual = "";
    private String eventDate = "";
    private String eventTime = "";
    private String submitHost = "";
    private String numExecHosts = "";
    private String jobStatus = "";
    
    /**
     * The actual value of hostFactor, which is a string taken from the log file.
     */
    private double factor = 1.0;

    /**
     * Creates a new event log file.
     */
    EventLogParser(File file, String timezone, String submitHost, boolean useScalingFactor) {
        super(file.getPath());
        m_eventRecords = new Vector();
        m_submitHost = submitHost;
        m_file = file;

        m_useScalingFactor = useScalingFactor;

        m_timezone = timezone;

    }

    public void openLogFile() {
        try {
            FileInputStream fileInputStream = new FileInputStream(m_file);

            if (getName().matches("^.*\\.gz$")) {
                m_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(fileInputStream)));
            } else {
                m_reader = new BufferedReader(new InputStreamReader(fileInputStream));
            }
        } catch (Exception e) {
            Log.warn(LOG, "Exception opening file " + m_file.getPath(), e);
        }
    }

    /**
     * Returns true if the file name matches today's date (ie the log
     * is currently being used).
     */
    boolean isActiveLog() {
        /*
         * D.Kant -- Comment out original code segments --
         * Date currentDate = Calendar.getInstance().getTime();
         * SimpleDateFormat dateFormatter = new SimpleDateFormat(EVENT_DATE_FORMAT);
         * dateFormatter.setTimeZone(TimeZone.getTimeZone(UTC));
         * String formattedDate = dateFormatter.format(currentDate);
         * D.Kant -- Begin new code segments --
         */
        SimpleDateFormat m_dateFormatter = new SimpleDateFormat(EVENT_DATE_FORMAT);
        m_dateFormatter.setTimeZone(TimeZone.getDefault());

        Date currentDate = Calendar.getInstance().getTime();
        String formattedDate = m_dateFormatter.format(currentDate);

        /*
         * D.Kant -- End --
         */
        String EventLogPrefix = "lsb.acct.";
        String CurrentLog = EventLogPrefix + formattedDate;

        return CurrentLog.equals(getName());
    }

    /**
     * Parses the log file and returns an iterator over all the
     * event records found within the file.
     */
    public void parse() {
        Log.info(LOG, "Processing " + getFullPath());

        String line = null;

        /**
         * pattern to obtain the LSF version
         */
        String pattern_init = "^\"JOB_FINISH\"" + //Event Type
                "\\s" + // Space separating different fields
                "\"(.*?)\"" + // Version number
                "\\s.*$";

        m_delimiter_init = Pattern.compile(pattern_init);

        Vector<String> fields = new Vector<String>();

        try {

            while ((line = m_reader.readLine()) != null) {

                line = line.trim();
                String lsf_format = "";
                // Getting the LSF version from each line
                if (m_delimiter_init.matcher(line).matches()) {


                    Matcher matcher = m_delimiter_init.matcher(line);
                    while (matcher.find()) {
                        lsf_format = matcher.group(1);
                    }
                }

                createPattern(lsf_format);
                fields.clear();

                if (m_delimiter.matcher(line).matches()) {


                    Matcher matcher = m_delimiter.matcher(line);

                    while (matcher.find()) {

                        for (int j = 1; j <= matcher.groupCount(); j++) {
                            fields.add(matcher.group(j));
                        }
                    }

                    setFields(fields);

                    //////////////////////////////////////////////////////////////////////////////////
                    // Reject ALL records that have EXIT return code AND do not consume CPU resources
                    // e.g. Jobs which are terminated while pending
                    //////////////////////////////////////////////////////////////////////////////////
                    if (!rejectRecordNoResourcesConsumed()) {
                        try {
                            EventRecord record = createEventRecord();
                            m_eventRecords.add(record);
                        } catch (Exception ae) {
                            Log.warn(LOG, "Unable to create record, skipping: " + ae.getMessage());
                        }
                    }
                }

            }
        } catch (IOException e) {
            Log.warn(LOG, "Exception reading file " + getName(), e);
        } catch (Exception e) {
            Log.warn(LOG, "Unable to create record, skipping: " + e.getMessage());
        }
    }

    /**
     * Useful for junit
     */
    public Vector getEventRecords() {
        return m_eventRecords;
    }

    /**
     * Returns a new EventRecord derived from the line of data
     * (parsed from the event log).
     */
    EventRecord createEventRecord() {

        EventRecord record = new EventRecord();
        if (m_submitHost == null || m_submitHost.equals("") || m_submitHost.equals("CE_HOSTNAME")) {
            record.setSubmitHost(submitHost);
        } else {
            record.setSubmitHost(m_submitHost);
        }
        record.setJobName(jobName);

        
        double cpu = Double.valueOf(cpuDuration);
        
        /******************************************************************
         * Savannah bug #85096:  Only scale the CPU and wallclock times 
         * if it is set in the configuration file.
         ******************************************************************/
        
        if (m_useScalingFactor) {
        	factor = Double.valueOf(hostFactor);
        }
        
        // factor is 1.0 by default
    	cpu = cpu * factor;
        

        // Get CPU time as a String representing an integer.
        String cpuS = Long.toString(Math.round(cpu));
        record.setCpuDurationSeconds(cpuS);

        /*
         * Convert m_cpuDurationSeconds into hh::mm:ss
         */
        String cpuD_string = secondsToString(cpu);
        TimeInterval interval = new TimeInterval(cpuD_string);
        record.setCpuDuration(interval.getIso8601());

        /*
         * Set localUserGroup (e.g. "dteam001" becomes dteam)
         */
        String localUserGroup = LocalUserID.replaceAll("[\"\\d{3}]", "");
        record.setLocalUserGroup(localUserGroup);

        /*
         * Set localUserID
         */
        String localUserID = LocalUserID.replaceAll("\"", "");
        record.setLocalUserID(localUserID);

        record.setMemoryReal(memoryReal);
        record.setMemoryVirtual(memoryVirtual);

        setTimeValues(record);

        return record;
    }

    /**
     * Sets all time related values of the EventRecord (derived from
     * the list).
     */
    private void setTimeValues(EventRecord record) {

        TimeInterval interval;

        /*
         * Set wallDurationSeconds
         */
        Double dispatch_time = new Double(startTime);
        Double event_time = new Double(eventDate);
        Double wds_double = new Double(event_time.doubleValue() - dispatch_time.doubleValue());
        
        // factor is 1.0 by default; if the option m_useScalingFactor is set, it will have
        // been set in createEventRecord()
        Double wds_norm = wds_double * factor;
        record.setWallDurationSeconds(Integer.toString(wds_norm.intValue()));

        /*
         * Convert wallDurationSeconds into hh::mm:ss
         */
        String wds_string = secondsToString(wds_norm.doubleValue());
        interval = new TimeInterval(wds_string);
        record.setWallDuration(interval.getIso8601());

        /*
         * lsf start/stop times defined in epoch seconds which must be adjusted into UTC format.
         */
        TimeZone timezone = TimeZone.getTimeZone(m_timezone);
        Time time = new Time(eventDate, timezone);
        record.setStopTimeEpoch(time.getEpoch());
        record.setStopTime(time.getIso8601());

        String stopTimeUtc = time.getIso8601Utc();
        record.setStopTimeUtc(stopTimeUtc);
        timezone = time.getTimeZone();
        time = new Time(startTime, timezone);
        record.setStartTimeEpoch(time.getEpoch());
        record.setStartTime(time.getIso8601());
        record.setStartTimeUtc(time.getIso8601Utc());
        record.setEventDate(stopTimeUtc.substring(0, 10));
        record.setEventTime(stopTimeUtc.substring(11, 19));
    }

    /*
     * Adjusts the offset used to read attributes from the array
     * if certain elements contain '0' (ie are empty).
     */
    private int getOffset(Vector<String> fields) {
        String field = (String) fields.get(22);
        int offSet = 0;

        if (!field.equals("0") && !fields.get(24).equals("0")) {
            offSet = 0;
        }

        if (field.equals("0") && !fields.get(23).equals("0")) {
            offSet = 1;
        }

        if (field.equals("0") && fields.get(23).equals("0")) {
            offSet = 2;
        }

        if (!field.equals("0") && fields.get(24).equals("0")) {
            offSet = 1;
        }

        return offSet;
    }

    /**
     * Closes down the log file object. This is called once all meaningful
     * data has been read from the file.
     */
    public void close() {
        if (m_reader != null) {
            try {
                m_reader.close();
            } catch (IOException e) {
                Log.warn(LOG, "Exception closing file " + getName(), e);
            }
        }
    }

    /**
     * Return canonical path, with built-in exception handling.
     * If an exception is thrown only the relative path is returned.
     */
    public String getFullPath() {
        String path;

        try {
            path = getCanonicalPath();
        } catch (Exception e) {
            Log.warn(LOG, "Exception trying to get filename", e);
            path = getName();
        }

        return path;
    }

    /**
     * Returns an iterator of all the parsed event records.
     */
    public EventRecordIterator getRecords() {
        return new EventRecordIterator(m_eventRecords.iterator());
    }

    /**
     * This method checks that certain fileds have the correct type.
     * Reject records in cases where the regular expression does not correctly
     * match against the fields.
     */
    private boolean rejectRecord(Vector fields, int offSet) {
        boolean valField = false;
        /*
         * Check exit status of job
         * Savannah Bug # 26100
         */
        if (((String) fields.get(26 - offSet)).equals(JOB_EXIT_STATUS)) {
            //valField = true;
        } else {
            // Check that system and user time match a float or integer
            Pattern p = Pattern.compile("[-+]?([0-9]*\\.)?[0-9]+");
            Matcher m1 = p.matcher((String) fields.get(30 - offSet));
            Matcher m2 = p.matcher((String) fields.get(31 - offSet));

            if (!m1.matches() || !m2.matches()) {
                valField = true;
            }
        }

        return valField;
    }

    /**
     * Simple mechanism to reject record based on the "Number of processors used for execution"
     */
    private boolean rejectRecordNoResourcesConsumed() {
        // default: reject record
        boolean retVal = false;

        /////////////////////////////////////////////////////////////////////
        // Criteria to Reject Record
        // Must have an EXIT for jobExitCode and Consume No resources
        // //////////////////////////////////////////////////////////////////
        if (jobStatus.equals(JOB_EXIT_STATUS) && numExecHosts.equals("0")) {
            retVal = true;
        }

        return retVal;

    }

    /**
     * Converts time in seconds to a <code>String</code> in the
     * format HH:mm:ss.
     */
    private String secondsToString(double time) {
        int seconds = (int) (time % 60);
        int minutes = (int) ((time / 60) % 60);
        int hours = (int) ((time / 3600) % 24);
        int days = (int) ((time / 3600) / 24);
        hours += (days * 24);

        String secondsStr = ((seconds < 10) ? "0" : "") + seconds;
        String minutesStr = ((minutes < 10) ? "0" : "") + minutes;
        String hoursStr = ((hours < 10) ? "0" : "") + hours;

        return new String(hoursStr + ":" + minutesStr + ":" + secondsStr);
    }

    /**
     * Puts the values from the different fields in the log file into the
     * correct variables
     */
    private void setFields(Vector<String> fields) {

        eventDate = fields.get(0);
        eventTime = fields.get(0);
        jobName = fields.get(1);
        LocalUserID = fields.get(3);
        startTime = fields.get(2);
        submitHost = fields.get(4);
        hostFactor = fields.get(8);
        float utime = Float.parseFloat(fields.get(9));
        float stime = Float.parseFloat(fields.get(10));
        float cpu = utime + stime;
        cpuDuration = String.valueOf(cpu);
        memoryReal = fields.get(18);
        memoryVirtual = fields.get(19);
        stopTime = fields.get(0);
        numExecHosts = fields.get(5);
        jobStatus = fields.get(7);
    }

    /** Creates the pattern used to match every line of the lsf log file
     * that corresponds to one EventRecord
     */
    private void createPattern(String lsf_format) {

        String lsfRusage_pattern = "";
        for (int c = 0; c <= 18; c++) {
            if (c == 0 || c == 1) {
                lsfRusage_pattern = lsfRusage_pattern + "([-0-9.]*?)\\s";
            } else {
                lsfRusage_pattern = lsfRusage_pattern + "[-0-9.]*?\\s";
            }
        }

        String pattern = "^\"JOB_FINISH\"" + //Event Type
                "\\s" + // Space separating different fields
                "\".*?\"" + // Version number
                "\\s"
                + "(\\d*)" + // EventTime
                "\\s"
                + "(\\d*)" + // JobID
                "\\s"
                + "\\d*" + // userId
                "\\s"
                + "\\d*" + // options
                "\\s"
                + "\\d*" + // numProcessors
                "\\s"
                + "\\d*" + // submitTime
                "\\s"
                + "\\d*" + // beginTime
                "\\s"
                + "\\d*" + // termTime
                "\\s"
                + "(\\d*)" + // startTime
                "\\s"
                + "\"(.*?)\"" + // userName
                "\\s"
                + "\".*?\"" + // queue
                "\\s"
                + "\".*?\"" + // resReq
                "\\s"
                + "\".*?\"" + // dependCond
                "\\s"
                + "\".*?\"" + // preExecCmd
                "\\s"
                + "\"(.*?)\"" + // fromHost
                "\\s"
                + "\".*?\"" + // cwd
                "\\s"
                + "\".*?\"" + // inFile
                "\\s"
                + "\".*?\"" + // outFile
                "\\s"
                + "\".*?\"" + // errFile
                "\\s"
                + "\".*?\"" + // jobFile
                "\\s"
                + "(?:0\\s(\\d*)\\s\".*\"\\s|\\d+\\s\".*\"\\s(\\d+)\\s\".*\"\\s)" + //numAskedHosts + askedHosts
                "(\\d*)" + // jStatus
                "\\s"
                + "(\\d*\\.\\d)" + // hostFactor
                "\\s"
                + ".*\\s"
                + lsfRusage_pattern + // The 19 fields on the lsfRusage section
                "\"(.*?)\"" + // mailUser
                "\\s"
                + "\"(.*?)\"" + // projectName
                "\\s"
                + "(\\d*|-1)" + // exitStatus
                "\\s"
                + "(\\d*|-1)" + // maxNumProcessors
                "\\s"
                + "\"(.*?)\"" + // loginShell
                "\\s"
                + "\"(.*?)\"" + // timeEvent
                "\\s"
                + "(\\d*|-1)" + // idx
                "\\s"
                + "(\\d*|-1)" + // maxRMem
                "\\s"
                + "(\\d*|-1)" + // maxRSwap
                "\\s"
                + "\"(.*?)\"" + // inFileSpool
                "\\s"
                + "\"(.*?)\"" + // commandSpool
                "\\s"
                + "\"(.*?)\"";  // srvId


        if (!lsf_format.startsWith("5")) {
            pattern = pattern + "\\s"
                    + "\"(.*?)\"" + // sla
                    "\\s"
                    + "(\\d*|-1)" + // exceptMask
                    "\\s"
                    + "\"(.*?)\"" + // additionalInfo
                    "\\s"
                    + "(\\d*|-1)" + // exitInfo
                    "\\s"
                    + "\"(.*?)\"" + // warningAction
                    "\\s"
                    + "(\\d*|-1)" + // warningTimePeriod
                    "\\s"
                    + "\"(.*?)\"" + // chargedSAAP
                    "\\s"
                    + "\"(.*?)\"$";  // licenseProject

        }

        m_delimiter = Pattern.compile(pattern);


    }
}
